/*create by user 02/06/2021 Jacobian error*/
#include "g2o_opt_type.h"
const double SMALL_EPS = 1e-5;

Eigen::Vector3f QuatPlane::log(const QuatPlane & other)
{
    double theta;
    return logAndTheta(other, &theta);
}

Eigen::Vector3f QuatPlane::logAndTheta(const QuatPlane & other, double * theta)
{

    double n = other._plane.normalized().vec().norm();
    double w = other._plane.normalized().w();
    double squared_w = w*w;

    double two_atan_nbyw_by_n;
    // Atan-based log thanks to
    //
    // C. Hertzberg et al.:
    // "Integrating Generic Sensor Fusion Algorithms with Sound State
    // Representation through Encapsulation of Manifolds"
    // Information Fusion, 2011

    if (n < SMALL_EPS)
    {
        // If quaternion is normalized and n=1, then w should be 1;
        // w=0 should never happen here!
        assert(fabs(w)>SMALL_EPS);

        two_atan_nbyw_by_n = 2./w - 2.*(n*n)/(w*squared_w);
    }
    else
    {
        if (fabs(w)<SMALL_EPS)
        {
            if (w>0)
            {
                two_atan_nbyw_by_n = M_PI/n;
            }
            else
            {
                two_atan_nbyw_by_n = -M_PI/n;
            }
        }
        two_atan_nbyw_by_n = 2*atan(n/w)/n;
    }

    *theta = two_atan_nbyw_by_n*n;
    return two_atan_nbyw_by_n * other._plane.normalized().vec();
}
/*
Eigen::Vector3d  QuatPlane::log_map(const QuatPlane q){
    Eigen::Vector3d qv;
    qv[0] = q._plane.x();
    qv[1] = q._plane.y();
    qv[2] = q._plane.z();
    double qw = q._plane.w();
    Eigen::Vector3d res = qv*std::atan(qv.norm()/qw)/(qv.norm()+1e-5);
    return res;

}*/

QuatPlane QuatPlane::exp(const Eigen::Vector3f & omega)
{
    double theta;
    return expAndTheta(omega, &theta);
}

QuatPlane QuatPlane::expAndTheta(const Eigen::Vector3f & omega, double * theta)
{
    *theta = omega.norm();
    double half_theta = 0.5*(*theta);

    double imag_factor;
    double real_factor = cos(half_theta);
    if((*theta)<SMALL_EPS)
    {
        double theta_sq = (*theta)*(*theta);
        double theta_po4 = theta_sq*theta_sq;
        imag_factor = 0.5-0.0208333*theta_sq+0.000260417*theta_po4;
    }
    else
    {
        double sin_half_theta = sin(half_theta);
        imag_factor = sin_half_theta/(*theta);
    }

    return QuatPlane(Eigen::Quaterniond(real_factor,
                           imag_factor*omega.x(),
                           imag_factor*omega.y(),
                           imag_factor*omega.z()));
}

// right jacobian of SO(3)
Eigen::Matrix3f QuatPlane::JacobianR(const Eigen::Vector3f& w)
{
    Eigen::Matrix3f Jr = Eigen::Matrix3d::Identity();
    double theta = w.norm();
    if(theta<0.00001)
    {
        return Jr;// = Matrix3d::Identity();
    }
    else
    {
        Eigen::Vector3f k = w.normalized();  // k - unit direction vector of w
        Eigen::Matrix3f K = g2o::hat(k);
        Jr =   Eigen::Matrix3d::Identity()
               - (1-cos(theta))/theta*K
               + (1-sin(theta)/theta)*K*K;
    }
    return Jr;
}
Eigen::Matrix3f QuatPlane::JacobianRInv(const Eigen::Vector3f& w)
{
    Eigen::Matrix3f Jrinv = Eigen::Matrix3f::Identity();
    double theta = w.norm();

    // very small angle
    if(theta < 0.00001)
    {
        return Jrinv;
    }
    else
    {
        Eigen::Vector3f k = w.normalized();  // k - unit direction vector of w
        Eigen::Matrix3f K = g2o::hat(k);
        Jrinv = Eigen::Matrix3f::Identity()
                + 0.5*g2o::hat(w)
                + ( 1.0 - (1.0+cos(theta))*theta / (2.0*sin(theta)) ) *K*K;
    }

    return Jrinv;
    /*
 * in gtsam:
 *
 *   double theta2 = omega.dot(omega);
 *  if (theta2 <= std::numeric_limits<double>::epsilon()) return I_3x3;
 *  double theta = std::sqrt(theta2);  // rotation angle
 *  * Right Jacobian for Log map in SO(3) - equation (10.86) and following equations in
 *   * G.S. Chirikjian, "Stochastic Models, Information Theory, and Lie Groups", Volume 2, 2008.
 *   * logmap( Rhat * expmap(omega) ) \approx logmap( Rhat ) + Jrinv * omega
 *   * where Jrinv = LogmapDerivative(omega);
 *   * This maps a perturbation on the manifold (expmap(omega))
 *   * to a perturbation in the tangent space (Jrinv * omega)
 *
 *  const Matrix3 W = skewSymmetric(omega); // element of Lie algebra so(3): W = omega^
 *  return I_3x3 + 0.5 * W +
 *         (1 / (theta * theta) - (1 + cos(theta)) / (2 * theta * sin(theta))) *
 *             W * W;
 *
 * */
}
/*
Eigen::Vector3d  QuatPlane::log_map(const QuatPlane q){
    Eigen::Vector3d qv;
    qv[0] = q._plane.x();
    qv[1] = q._plane.y();
    qv[2] = q._plane.z();
    double qw = q._plane.w();
    Eigen::Vector3d res = qv*std::atan(qv.norm()/qw)/(qv.norm()+1e-5);
    return res;

}
 */
// left jacobian of SO(3), Jl(x) = Jr(-x)
Eigen::Matrix3f QuatPlane::JacobianL(const Eigen::Vector3f& w)
{
    Eigen::Matrix3f Jl = Eigen::Matrix3f::Identity();
    Eigen::Matrix3f Jl_old = JacobianR(-w); // original version

    double theta = w.norm();
    if(theta < 0.00001)
    {
        return Jl;
    }
    else
    {
        Eigen::Vector3f k = w.normalized();  // k - unit direction vector of w
        Eigen::Matrix3f K_hat = g2o::hat(k);
        Jl = (sin(theta) / theta ) * Eigen::Matrix3f::Identity()
             + (1 - sin(theta)/theta) * k * k.transpose()
             + ((1 - cos(theta)) / theta) * K_hat;
    }

    std::cout << "Jl old : " << Jl_old << std::endl;
    std::cout << "Jl  : " << Jl << std::endl;

    return Jl;

    // return JacobianR(-w);
}
// left jacobian inverse
Eigen::Matrix3f QuatPlane::JacobianLInv(const Eigen::Vector3f& w)
{
    Eigen::Matrix3f JlInv_old = JacobianRInv(-w); // original version
    Eigen::Matrix3f JlInv = Eigen::Matrix3f::Identity();

    double theta = w.norm();
    if(theta < 0.00001)
    {
        return JlInv;
    }
    else
    {
        Eigen::Vector3f k = w.normalized();  // k - unit direction vector of w
        Eigen::Matrix3f K_hat = g2o::hat(k);
        double half_theta = theta / 2;
        double cot_half_theta = cos(half_theta) / sin(half_theta);
        JlInv = half_theta * cot_half_theta * Eigen::Matrix3f::Identity()
                + (1 - half_theta * cot_half_theta) * k * k.transpose()
                - half_theta * K_hat;
    }

//        std::cout << "JlInv old : " << JlInv_old << std::endl;
//        std::cout << "JlInv  : " << JlInv << std::endl;

    return JlInv;

    // return JacobianRInv(-w);
}

namespace g2o
{
void EdgeSE3ProjectPlane::computeError()
{
    const VertexSE3Expmap* vj = static_cast<const VertexSE3Expmap*>(_vertices[1]);
    g2o::SE3Quat T(vj->estimate());
    Eigen::Matrix4f  T_w_c= T.to_homogeneous_matrix().transpose();
    const VertexQuatPlane* vi = static_cast<const VertexQuatPlane*>(_vertices[0]);
    QuatPlane plane_w = vi->estimate();
    Eigen::Quaterniond q_w = plane_w._plane;
    Eigen::Vector4f  plane_w_homo = plane_w.HesseToHomogenous(plane_w._normal,plane_w._dis);
    // Eigen::Vector4d plane_c_home = T_w_c*plane_w_homo ;
    QuatPlane q_hat_w(plane_w_homo);
    Eigen::Matrix3d R = T.to_homogeneous_matrix().block(0,0,3,3);
    Eigen::Vector3d t = T.to_homogeneous_matrix().block(0,3,3,1);
    Quaterniond q;
    q.vec() = R.transpose()*_measurement._plane.vec();
    q.w() = t.transpose()*_measurement._plane.vec()+_measurement._plane.w();
    QuatPlane obs_w(q);//T_w_c*_measurement._plane);
    QuatPlane q_u(obs_w._plane*q_hat_w._plane.inverse());

    _error =  QuatPlane::log(q_u);
    //std::cout<<"error is : "<<_error.transpose()<<std::endl;
    int a = 0;
}

void EdgeSE3ProjectPlane::linearizeOplus()
{
    // order
    // vertice of camera pose
    VertexSE3Expmap* vj = static_cast<VertexSE3Expmap*>(_vertices[1]);
    g2o::SE3Quat T(vj->estimate());

    // vertice of 3d plane
    VertexQuatPlane * vi = static_cast<VertexQuatPlane *>(_vertices[0]);
    QuatPlane plane_w = vi->estimate();//vi vector3d
    //std::cout << "edge plane quaternion: " << plane_w._plane.coeffs() << std::endl;//plane.unit_quaternion()
    Eigen::Quaterniond plane_w_unit_quaternion = plane_w._plane.normalized();
    Eigen::Matrix3d R = T.to_homogeneous_matrix().block(0,0,3,3);
    Eigen::Vector3d t = T.to_homogeneous_matrix().block(0,3,3,1);
    Eigen::Quaterniond _measurement_w_q;
    _measurement_w_q.vec() = R.transpose() * _measurement._plane.vec();
    _measurement_w_q.w() = t.transpose() * _measurement._plane.vec() + _measurement._plane.w();
    QuatPlane gt_plane_w(_measurement_w_q);
    QuatPlane distance_bet_plane(gt_plane_w._plane*plane_w._plane.inverse());//  we use measurement as ground truth.
    Eigen::Vector3d distance_bet_plane_log = QuatPlane::log(distance_bet_plane);
    _jacobianOplusXi = -1 * QuatPlane::JacobianRInv(distance_bet_plane_log); // -Jr(u)^(-1)
    Eigen::Quaterniond phi=gt_plane_w._plane;

    Eigen::Vector3d phi_v = phi.vec();
    double phi_n = phi.norm();
    double phi_n_3 = phi_n*phi_n*phi_n;
    double phi_v_norm = phi_v.norm();
    double phi_q4 = phi.w();

    Eigen::Vector4d y;
    y.head(3)=phi_v;y(3) = phi_q4;
    QuatPlane phi_p(phi);
    Eigen::Vector3d u3 = QuatPlane::log(phi_p);
    Eigen::Vector3d w3 = QuatPlane::log(plane_w_unit_quaternion);
    // jacobian_error_u
    Eigen::Matrix<double,3,3> jacobian_error_u;
    QuatPlane beta_exp(QuatPlane::exp(u3)._plane*QuatPlane::exp(w3)._plane.inverse());
    Eigen::Vector3d beta_log = QuatPlane::log(beta_exp);/**************************/
    jacobian_error_u = QuatPlane::JacobianLInv(beta_log);
    // jacobian_u_phi
    Eigen::Matrix<double, 3, 4> jacobian_u_phi;/*****name**/
    jacobian_u_phi.block<3, 3>(0, 0) = 2 * acos(phi_q4) * (Eigen::Matrix3d::Identity() / phi_v_norm -
                                                        phi_v * phi_v.transpose() / sqrt(phi_n_3));
    jacobian_u_phi.block<3, 1>(0, 3) = -1 * 2.0 / (phi_v_norm * sqrt(1.0 - phi_q4 * phi_q4 + 0.0001)) * phi_v;
    // jacobian_phi_y   /********************/
    Eigen::Matrix<double, 4, 4> jacobian_phi_y;/*no homo    use quat*/
    Eigen::Vector4d _measurement_w_vec;_measurement_w_vec << _measurement_w_q.x(),_measurement_w_q.y(),_measurement_w_q.z(),_measurement_w_q.w();
    double y_norm = _measurement_w_vec.norm();
    double y_norm_3 = y_norm*y_norm*y_norm;
    jacobian_phi_y = Eigen::Matrix4d::Identity() / y_norm - _measurement_w_vec * _measurement_w_vec.transpose()/ std::sqrt(y_norm_3);
    // jacobian_y_pose
    Eigen::Vector3d pi_x_v;//****//
    pi_x_v = plane_w_unit_quaternion.vec();//x(), plane_w_unit_quaternion.y(), plane_w_unit_quaternion.z();
    Eigen::Matrix<double, 4, 6> jacobian_y_pose = Eigen::Matrix<double, 4, 6>::Zero();
    jacobian_y_pose.block<3, 3>(0, 0) = T.rotation().toRotationMatrix().transpose() * g2o::hat(pi_x_v);
    jacobian_y_pose.block<1, 3>(3, 0) = T.translation().transpose() * g2o::hat(pi_x_v);
    jacobian_y_pose.block<1, 3>(3, 3) = pi_x_v.transpose();
    _jacobianOplusXj = jacobian_error_u * jacobian_u_phi * jacobian_phi_y * jacobian_y_pose; // 3 x 6 matrix  J_l_u_inv
    //std::cout << "_jacobianOplusXj has been done " << std::endl;
}

}